"""The idea behind this module comes from the conjunction of two concepts:

- [1] [Classifier Two-Sample Test](https://arxiv.org/abs/1610.06545)
- [2] [Recursive Feature Elimination](\
    https://scikit-learn.org/stable/modules/generated/\
    sklearn.feature_selection.RFE.html)

In [1] classifier performances are used to determine how similar two samples are. More
specifically, imagine to have two samples: `reference` and `test`. In order to assess
whether `reference` and `test` have been drawn from the same distribution, we could
train a classifier in classifying which instances belong to which sample. If the
model easily distinguishes instances from the two samples, then the two samples
have been probably drawn from two different distributions. Conversely, if the
classifier struggles to distinguish them, then it is likely that the samples have
been drawn from the same distribution.

In the context of drift detection, the classifier two-sample test can be used to
assess whether drift has happened between the reference and the test set and to
which degree.

The classes of this module take this idea one step further and attempt
to reduce the drift using recursive feature selection. After a classifier
is trained to distinguish between `reference` and `test`, the feature
importance of the classifier is used to determine which features contribute
the most in distinguishing between the two sets. The most important features
are then eliminated and the procedure is repeated until the classifier is not
able anymore to distinguish between the two samples, or until a certain amount
of features has been removed.

This module contains the following classes:
- `SampleSimilarityDriftRFE`: base class for drift-based sample similarity
    feature selection.
"""

from collections import defaultdict
from numbers import Integral

import numpy as np
import pandas as pd
from joblib import Parallel, delayed, effective_n_jobs
from sklearn.base import ClassifierMixin, clone
from sklearn.metrics import check_scoring
from sklearn.model_selection import check_cv
from sklearn.utils._tags import _safe_tags, get_tags
from sklearn.utils.validation import validate_data

from felimination.importance import PermutationImportance
from felimination.rfe import FeliminationRFECV, _train_score_get_importance


class SampleSimilarityDriftRFE(FeliminationRFECV):
    """Recursively discards the features that introduce the highest drift.

    The algorithm of feature selection goes as follows:
    ```
    Split X into two sets using the `split_column`: X1 and X2
    create target array y1 for X1 as an array of zeroes
    create target array y2 for X2 as an array of ones
    vertically concatenate X1, X2 and y1 and y2, obtaining X_ss and y_ss
    Calculate Cross-validation performances of the estimator on X_ss and y_ss.
    while cross-validation-performances > max_score and n_features > min_features_to_select:
        Discard most important features
        Calculate Cross-validation performances of the estimator on X_ss and y_ss using the new feature set.
    ```

    Parameters
    ----------
    clf : ``Classifier`` instance
        A Classifier with a ``fit`` method.
    step : int or float, default=1
        If greater than or equal to 1, then ``step`` corresponds to the
        (integer) number of features to remove at each iteration.
        If within (0.0, 1.0), then ``step`` corresponds to the percentage
        (rounded down) of **remaining** features to remove at each iteration.
        Note that the last iteration may remove fewer than ``step`` features in
        order to reach ``min_features_to_select``.
    max_score : float, default=0.55
        Stops the feature selection procedure when the
        cross-validation score of the sample similarity classifier is
        lower than `max_score`.
    min_features_to_select : int or float, default=1
        The minimum number of features to select. If `None`, half of the features are
        selected. If integer, the parameter is the absolute number of features
        to select. If float between 0 and 1, it is the fraction of the features to
        select.
    split_column : str, default='split'
        The name of the column in the dataset that will be used to split the dataset
        into two sets.
    split_value : Any, default=None
        If defined, this value will be used to split the dataset into two sets.
    split_frac : float, default=0.5
        If split_value, split frac is used to determine a split_value. The split
        frac corresponds to the quantile of the split_column to use as the split_value.
    split_unique_values: bool, default=True
        Whether to calculate the quantile of the split_column to use as the split_value
        based on the unique values of the split_column.
    cv : int, cross-validation generator or an iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

            - None, to use the default 5-fold cross-validation,
            - integer, to specify the number of folds.
            - :term:`CV splitter`,
            - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`~sklearn.model_selection.StratifiedKFold` is used. If the
        estimator is a classifier or if ``y`` is neither binary nor multiclass,
        :class:`~sklearn.model_selection.KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.
    scoring : str, callable or None, default=None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.
    verbose : int, default=0
        Controls verbosity of output.
    n_jobs : int or None, default=None
        Number of cores to run in parallel while fitting across folds.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
    importance_getter : str or callable, default='auto'
        If 'auto', uses the feature importance either through a `coef_`
        or `feature_importances_` attributes of estimator.

        Also accepts a string that specifies an attribute name/path
        for extracting feature importance.
        For example, give `regressor_.coef_` in case of
        :class:`~sklearn.compose.TransformedTargetRegressor`  or
        `named_steps.clf.feature_importances_` in case of
        :class:`~sklearn.pipeline.Pipeline` with its last step named `clf`.

        If `callable`, overrides the default feature importance getter.
        The callable is passed with the fitted estimator and the validation set
        (X_val, y_val, estimator) and it should return importance for each feature.


    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        The classes labels.
    clf_ : ``Classifier`` instance
        The fitted classifier used to select features.
    cv_results_ : dict of ndarrays
        A dict with keys:
        n_features : ndarray of shape (n_subsets_of_features,)
            The number of features used at that step.
        split(k)_test_score : ndarray of shape (n_subsets_of_features,)
            The cross-validation scores across (k)th fold.
        mean_test_score : ndarray of shape (n_subsets_of_features,)
            Mean of scores over the folds.
        std_test_score : ndarray of shape (n_subsets_of_features,)
            Standard deviation of scores over the folds.
        split(k)_train_score : ndarray of shape (n_subsets_of_features,)
            The cross-validation scores across (k)th fold.
        mean_train_score : ndarray of shape (n_subsets_of_features,)
            Mean of scores over the folds.
        std_train_score : ndarray of shape (n_subsets_of_features,)
            Standard deviation of scores over the folds.
    n_features_ : int
        The number of selected features.
    n_features_in_ : int
        Number of features seen during :term:`fit`. Only defined if the
        underlying estimator exposes such an attribute when fit.
    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.
    ranking_ : ndarray of shape (n_features,)
        The feature ranking, such that ``ranking_[i]`` corresponds to the
        ranking position of the i-th feature. Selected (i.e., estimated
        best) features are assigned rank 1.
    support_ : ndarray of shape (n_features,)
        The mask of selected features.

    # TODO: Add example
    """

    def __init__(
        self,
        clf: ClassifierMixin,
        *,
        step=1,
        max_score=0.55,
        min_features_to_select=1,
        split_col=0,
        split_value=None,
        split_frac=0.5,
        split_unique_values=True,
        cv=None,
        scoring=None,
        random_state=None,
        verbose=0,
        n_jobs=None,
        importance_getter="auto",
    ) -> None:
        self.max_score = max_score
        self.split_col = split_col
        self.split_value = split_value
        self.split_unique_values = split_unique_values
        self.split_frac = split_frac
        self.clf = clf
        super().__init__(
            estimator=clf,
            min_features_to_select=min_features_to_select,
            step=step,
            cv=cv,
            scoring=scoring,
            random_state=random_state,
            verbose=verbose,
            n_jobs=n_jobs,
            importance_getter=importance_getter,
        )

    def _build_sample_similarity_x_y(self, X, split_col_values):
        X1, X2 = self._split_X(X, split_col_values)
        y1 = np.zeros((X1.shape[0],))
        y2 = np.ones((X2.shape[0],))
        if isinstance(X, np.ndarray):
            X = np.vstack((X1, X2))
        else:
            X = pd.concat([X1, X2])
        y = np.concatenate((y1, y2))
        return X, y

    def _split_X(self, X, split_col_values):
        if self.split_value is not None:
            split_value = self.split_value
        else:
            if self.split_unique_values:
                unique_split_col_values = np.unique(split_col_values)
                split_value = np.quantile(unique_split_col_values, self.split_frac)
            split_value = np.quantile(split_col_values, self.split_frac)

        mask = split_col_values >= split_value

        if isinstance(X, np.ndarray):
            X1 = X[mask]
            X2 = X[~mask]
        else:
            X1 = X.loc[mask]
            X2 = X.loc[~mask]
        return X1, X2

    def fit(self, X, y=None, groups=None, **fit_params):
        """Fit the RFE model and then the underlying clf on the selected features.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples.
        y : array-like of shape (n_samples,)
            The target values. Not used, kept for compatibility.
        groups : array-like of shape (n_samples,), default=None
            Group labels for the samples used while splitting the dataset into
            train/test set. Only used in conjunction with a "Group" :term:`cv`
            instance.
        **fit_params : dict
            Additional parameters passed to the `fit` method of the underlying
            clf.

        Returns
        -------
        self : object
            Fitted selector.
        """
        self._validate_params()
        validate_data(
            self,
            X,
            y,
            accept_sparse="csc",
            ensure_min_features=2,
            ensure_all_finite=not get_tags(self.estimator).input_tags.allow_nan,
            dtype=None,
        )

        if isinstance(self.split_col, str):
            split_col_idx = list(self.feature_names_in_).index(self.split_col)
        else:
            split_col_idx = self.split_col

        if isinstance(X, np.ndarray):
            split_col_values = X[:, split_col_idx]
        else:
            split_col_name = X.columns[split_col_idx]
            split_col_values = X[split_col_name].values

        X, y = self._build_sample_similarity_x_y(X, split_col_values=split_col_values)

        # Initialization
        cv = check_cv(self.cv, y, classifier=True)
        scorer = check_scoring(self.clf, scoring=self.scoring)
        n_features = X.shape[1]

        if self.min_features_to_select is None:
            min_features_to_select = n_features // 2
        elif isinstance(self.min_features_to_select, Integral):  # int
            min_features_to_select = self.min_features_to_select
        else:  # float
            min_features_to_select = int(n_features * self.min_features_to_select)

        support_ = np.ones(n_features, dtype=bool)
        support_[split_col_idx] = False
        ranking_ = np.ones(n_features, dtype=int)

        current_number_of_features = support_.sum()
        self.cv_results_ = defaultdict(list)

        if self.verbose > 0:
            print("Fitting clf with %d features." % current_number_of_features)

        # Train model, score it and get importances
        if effective_n_jobs(self.n_jobs) == 1:
            parallel, func = list, _train_score_get_importance
        else:
            parallel = Parallel(n_jobs=self.n_jobs)
            func = delayed(_train_score_get_importance)

        features = np.arange(n_features)[support_]

        X_remaining_features, features = self._select_X_with_remaining_features(
            X, support=support_, n_features=n_features
        )

        scores_importances = parallel(
            func(
                self.clf,
                X_remaining_features,
                y,
                train,
                test,
                scorer,
                self.importance_getter,
            )
            for train, test in cv.split(X_remaining_features, y, groups)
        )

        test_scores_per_fold = [
            score_importance[1] for score_importance in scores_importances
        ]
        train_scores_per_fold = [
            score_importance[0] for score_importance in scores_importances
        ]

        # Update cv scores
        for train_or_test, scores_per_fold in zip(
            ["train", "test"], [train_scores_per_fold, test_scores_per_fold]
        ):
            for i, score in enumerate(scores_per_fold):
                self.cv_results_[f"split{i}_{train_or_test}_score"].append(score)
            self.cv_results_[f"mean_{train_or_test}_score"].append(
                np.mean(scores_per_fold)
            )
            self.cv_results_[f"std_{train_or_test}_score"].append(
                np.std(scores_per_fold)
            )
        self.cv_results_["n_features"].append(current_number_of_features)

        # Elimination
        while (
            np.mean(test_scores_per_fold) > self.max_score
            and current_number_of_features > min_features_to_select
        ):
            features = np.arange(n_features)[support_]
            if 0.0 < self.step < 1.0:
                step = int(max(1, self.step * current_number_of_features))
            else:
                step = int(self.step)
            # Eliminate most important features
            threshold = min(step, current_number_of_features - min_features_to_select)
            cv_importances = [
                score_importance[2] for score_importance in scores_importances
            ]
            mean_importances = np.mean(np.vstack(cv_importances), axis=0)
            ranks = np.argsort(-mean_importances)
            ranks = np.ravel(ranks)
            support_[features[ranks][:threshold]] = False
            ranking_[np.logical_not(support_)] += 1
            current_number_of_features = np.sum(support_)
            # Select remaining features
            features = np.arange(n_features)[support_]
            X_remaining_features, features = self._select_X_with_remaining_features(
                X, support=support_, n_features=n_features
            )

            if self.verbose > 0:
                print("Fitting clf with %d features." % current_number_of_features)

            # Train model, score it and get importances
            if effective_n_jobs(self.n_jobs) == 1:
                parallel, func = list, _train_score_get_importance
            else:
                parallel = Parallel(n_jobs=self.n_jobs)
                func = delayed(_train_score_get_importance)

            scores_importances = parallel(
                func(
                    self.clf,
                    X_remaining_features,
                    y,
                    train,
                    test,
                    scorer,
                    self.importance_getter,
                )
                for train, test in cv.split(X_remaining_features, y, groups)
            )
            train_scores_per_fold = [
                score_importance[0] for score_importance in scores_importances
            ]
            test_scores_per_fold = [
                score_importance[1] for score_importance in scores_importances
            ]

            # Update cv scores
            for train_or_test, scores_per_fold in zip(
                ["train", "test"], [train_scores_per_fold, test_scores_per_fold]
            ):
                for i, score in enumerate(scores_per_fold):
                    self.cv_results_[f"split{i}_{train_or_test}_score"].append(score)
                self.cv_results_[f"mean_{train_or_test}_score"].append(
                    np.mean(scores_per_fold)
                )
                self.cv_results_[f"std_{train_or_test}_score"].append(
                    np.std(scores_per_fold)
                )
            self.cv_results_["n_features"].append(current_number_of_features)

        features = np.arange(n_features)[support_]
        self.clf_ = clone(self.clf)
        X_remaining_features, features = self._select_X_with_remaining_features(
            X, support=support_, n_features=n_features
        )
        self.clf_.fit(X_remaining_features, y, **fit_params)

        self.n_features_ = support_.sum()
        self.support_ = support_
        self.ranking_ = ranking_
        self.cv_results_ = dict(self.cv_results_)
        return self

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        sub_estimator_tags = get_tags(self.estimator)
        tags.target_tags.required = False
        tags.input_tags.allow_nan = sub_estimator_tags.input_tags.allow_nan
        return tags


class PermImpSampleSimilarityDriftRFE(SampleSimilarityDriftRFE):
    """Preset of SampleSimilarityDriftRFE using permutation importance as importance getter.

    It has the following differences with RFECV from scikit-learn:

    - It supports an `importance_getter` function that also uses a validation
      set to compute the feature importances. This allows to use importance measures
      like permutation importance or shap.
    - Instead of using Cross Validation to select the number of features, it
      uses cross validation to get a more accurate estimate of the feature
      importances. This means that the number of features to select has to be
      set during initialization, similarly to RFE.
    - When `step` is a float value it is removes a percentage of the number
      of **remaining** features, not total like in RFE/RFECV. This allows to
      drop big chunks of feature at the beginning of the RFE process and to slow
      down towards the end of the process.
    - Has a plotting function
    - Adds information about the number of features selected at each step in the
      attribute `cv_results_`
    - Allows to change the number of features to be selected after fitting.

    Rater than that, it is a copy-paste of RFE, so credit goes to scikit-learn.

    The algorithm of feature selection goes as follows:
    ```
    while n_features > n_features_to_select:
        - The estimator is trained on the selected features and the score is
          computed using cross validation.
        - feature importance is computed for each validation fold on the validation
          set and then averaged.
        - The least important features are pruned.
        - The pruned features are removed from the dataset.
    ```

    Parameters
    ----------
    clf : ``Classifier`` instance
        A Classifier with a ``fit`` method.
    step : int or float, default=1
        If greater than or equal to 1, then ``step`` corresponds to the
        (integer) number of features to remove at each iteration.
        If within (0.0, 1.0), then ``step`` corresponds to the percentage
        (rounded down) of **remaining** features to remove at each iteration.
        Note that the last iteration may remove fewer than ``step`` features in
        order to reach ``min_features_to_select``.
    max_score : float, default=0.55
        Stops the feature selection procedure when the
        cross-validation score of the sample similarity classifier is
        lower than `max_score`.
    min_features_to_select : int or float, default=1
        The minimum number of features to select. If `None`, half of the features are
        selected. If integer, the parameter is the absolute number of features
        to select. If float between 0 and 1, it is the fraction of the features to
        select.
    split_column : str, default='split'
        The name of the column in the dataset that will be used to split the dataset
        into two sets.
    split_value : Any, default=None
        If defined, this value will be used to split the dataset into two sets.
    split_frac : float, default=0.5
        If split_value, split frac is used to determine a split_value. The split
        frac corresponds to the quantile of the split_column to use as the split_value.
    split_unique_values: bool, default=True
        Whether to calculate the quantile of the split_column to use as the split_value
        based on the unique values of the split_column.
    cv : int, cross-validation generator or an iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 5-fold cross-validation,
        - integer, to specify the number of folds.
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`~sklearn.model_selection.StratifiedKFold` is used. If the
        estimator is a classifier or if ``y`` is neither binary nor multiclass,
        :class:`~sklearn.model_selection.KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.
    scoring : str, callable or None, default=None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.
    verbose : int, default=0
        Controls verbosity of output.
    n_jobs : int or None, default=None
        Number of cores to run in parallel while fitting across folds.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
    n_repeats : int, default=5
        Number of times to permute a feature.
    random_state : int, RandomState instance, default=None
        Pseudo-random number generator to control the permutations of each
        feature.
        Pass an int to get reproducible results across function calls.
    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights used in scoring.
    max_samples : int or float, default=1.0
        The number of samples to draw from X to compute feature importance
        in each repeat (without replacement).
        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples.
        - If `max_samples` is equal to `1.0` or `X.shape[0]`, all samples
        will be used.
        While using this option may provide less accurate importance estimates,
        it keeps the method tractable when evaluating feature importance on
        large datasets. In combination with `n_repeats`, this allows to control
        the computational speed vs statistical accuracy trade-off of this method.


    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        The classes labels. Only available when `estimator` is a classifier.
    estimator_ : ``Estimator`` instance
        The fitted estimator used to select features.
    cv_results_ : dict of ndarrays
        A dict with keys:
        n_features : ndarray of shape (n_subsets_of_features,)
            The number of features used at that step.
        split(k)_test_score : ndarray of shape (n_subsets_of_features,)
            The cross-validation scores across (k)th fold.
        mean_test_score : ndarray of shape (n_subsets_of_features,)
            Mean of scores over the folds.
        std_test_score : ndarray of shape (n_subsets_of_features,)
            Standard deviation of scores over the folds.
        split(k)_train_score : ndarray of shape (n_subsets_of_features,)
            The cross-validation scores across (k)th fold.
        mean_train_score : ndarray of shape (n_subsets_of_features,)
            Mean of scores over the folds.
        std_train_score : ndarray of shape (n_subsets_of_features,)
            Standard deviation of scores over the folds.
    n_features_ : int
        The number of selected features.
    n_features_in_ : int
        Number of features seen during :term:`fit`. Only defined if the
        underlying estimator exposes such an attribute when fit.
    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.
    ranking_ : ndarray of shape (n_features,)
        The feature ranking, such that ``ranking_[i]`` corresponds to the
        ranking position of the i-th feature. Selected (i.e., estimated
        best) features are assigned rank 1.
    support_ : ndarray of shape (n_features,)
        The mask of selected features.
    """

    def __init__(
        self,
        clf: ClassifierMixin,
        *,
        step=1,
        max_score=0.55,
        min_features_to_select=1,
        split_col=0,
        split_value=None,
        split_frac=0.5,
        split_unique_values=True,
        cv=None,
        scoring=None,
        verbose=0,
        n_jobs=None,
        n_repeats=5,
        random_state=None,
        sample_weight=None,
        max_samples=1.0,
    ) -> None:
        self.n_repeats = n_repeats
        self.sample_weight = sample_weight
        self.max_samples = max_samples
        super().__init__(
            clf=clf,
            max_score=max_score,
            min_features_to_select=min_features_to_select,
            split_col=split_col,
            split_value=split_value,
            split_frac=split_frac,
            split_unique_values=split_unique_values,
            step=step,
            cv=cv,
            scoring=scoring,
            random_state=random_state,
            verbose=verbose,
            n_jobs=n_jobs,
            importance_getter=PermutationImportance(
                scoring=scoring,
                n_repeats=n_repeats,
                # Better not to do double parallelization
                n_jobs=1,
                random_state=random_state,
                sample_weight=sample_weight,
                max_samples=max_samples,
            ),
        )



"""This module contains the implementation of the Hybrid Genetic Algorithm-Importance with
Cross-Validation. The algorithm is implemented in the `HybridImportanceGACVFeatureSelector` class.
"""

from itertools import cycle, islice
from numbers import Integral, Real
from operator import itemgetter
from typing import Any, Iterable

import numpy as np
import pandas as pd
import seaborn as sns
from joblib import Parallel, delayed, effective_n_jobs
from sklearn.base import BaseEstimator, MetaEstimatorMixin, clone, is_classifier
from sklearn.feature_selection import SelectorMixin
from sklearn.linear_model._logistic import LogisticRegression
from sklearn.metrics import check_scoring
from sklearn.model_selection import check_cv
from sklearn.utils._param_validation import HasMethods, Interval, RealNotInt
from sklearn.utils._tags import get_tags
from sklearn.utils.metaestimators import available_if
from sklearn.utils.validation import check_is_fitted, validate_data

from felimination.importance import _train_score_get_importance


def _estimator_has(attr):
    """Check if we can delegate a method to the underlying estimator.

    First, we check the fitted `estimator_` if available, otherwise we check the
    unfitted `estimator`. We raise the original `AttributeError` if `attr` does
    not exist. This function is used together with `available_if`.
    """

    def check(self):
        if hasattr(self, "estimator_"):
            getattr(self.estimator_, attr)
        else:
            getattr(self.estimator, attr)

        return True

    return check


def _select_X_with_features(X, features):
    if isinstance(X, pd.DataFrame):
        return X[features]
    else:
        return X[:, features]


def _roundrobin(*iterables: Iterable[Iterable[Any]]) -> Iterable[Any]:
    """Implements the round robin algorithm.

    Given a some iterables extracts the first item from each iterable in turn,
    until all iterables are exhausted.

    For example roundrobin("ABC", "123") -> ["A", "1", "B", "2", "C", "3"]

    taken from itertools recipes
    https://docs.python.org/3.9/library/itertools.html#itertools-recipes

    Yields:
        Iterable[Any]: items from the passed iterables in a round robin fashion.
    """
    num_active = len(iterables)
    nexts = cycle(iter(it).__next__ for it in iterables)
    while num_active:
        try:
            for next_ in nexts:
                yield next_()
        except StopIteration:
            # Remove the iterator we just exhausted from the cycle.
            num_active -= 1
            nexts = cycle(islice(nexts, num_active))


def _deduplicate(seq: list):
    seen = set()
    return [x for x in seq if not (x in seen or seen.add(x))]


def rank_mean_test_score_overfit_fitness(pool):
    """Define the fitness function as the sum of the rank of the mean test score and the rank of the
    overfit.

    The rank of the mean test score is calculated by ranking the mean test score in ascending order.
    The rank of the overfit is calculated by ranking the overfit in ascending order.
    The overfit is calculated as the difference between the mean train score and the mean test score.
    The fitness is the sum of the rank of the mean test score and the rank of the overfit.

    Parameters
    ----------
    pool : list of dict
        Each element in the list is a dictionary with the following keys:
        - features: list of int
            The features selected for this element.
        - mean_test_score: float
            The mean test score of the element.
        - mean_train_score: float
            The mean train score of the element.

    Returns
    -------
    fitness : list of float
        The fitness of each element in the pool.
    """

    pool_df = pd.DataFrame(pool)
    pool_df["rank_mean_test_score"] = pool_df["mean_test_score"].rank(ascending=False)
    pool_df["overfit"] = pool_df["mean_train_score"] - pool_df["mean_test_score"]
    pool_df["rank_overfit"] = pool_df["overfit"].rank(ascending=True)
    pool_df["rank_sum"] = pool_df["rank_mean_test_score"] + pool_df["rank_overfit"]

    pool_df["rank_sum_rank"] = pool_df["rank_sum"].rank(ascending=False)
    return pool_df["rank_sum_rank"].to_list()


def rank_mean_test_score_fitness(pool):
    """Define the fitness function as the rank of the mean test score.

    The rank of the mean test score is calculated by ranking the mean test score in ascending order.

    Parameters
    ----------

    pool : list of dict
        Each element in the list is a dictionary with the following keys:
        - features: list of int
            The features selected for this element.
        - mean_test_score: float
            The mean test score of the element.
        - mean_train_score: float
            The mean train score of the element.

    Returns
    -------
    fitness : list of float
        The fitness of each element in the pool.
    """
    pool_df = pd.DataFrame(pool)
    pool_df["rank_mean_test_score"] = pool_df["mean_test_score"].rank(ascending=True)
    return pool_df["rank_mean_test_score"].to_list()


class HybridImportanceGACVFeatureSelector(
    SelectorMixin, MetaEstimatorMixin, BaseEstimator
):
    """Feature selection using Hybrid Genetic Algorithm-Importance with Cross-Validation.

    This feature selector uses a genetic algorithm to select features. The genetic algorithm
    is hybridized with feature importance. The feature importance is calculated using a
    cross-validation scheme. The algorithm works as follows:

    **Pool initialization:** The pool is initialized with random features. The number of features is
    randomly generated using a normal distribution with the average number of features to select and
    the standard deviation of the number of features to select as parameters. The number of features
    is clipped to be between the minimum number of features to select and the number of features in
    the dataset.

    **Cross Over:** The cross over is done by combining the features of the parents. The features
    are sorted by importance and the children are created by combining the features of the parents
    in a round-robin fashion. The number of features of the children is the average of the number of
    features of the parents. In this way, the children will have the most important features of the
    parents.

    **Mutation:** The mutation is done by randomly changing the number of features and replacing the
    least important features with random features.

    **Selection:** The selection is done by selecting the top `pool_size` solutions based on the
    fitness function.


    Parameters
    ----------
    estimator : object
        An estimator that follows the scikit-learn API and has a `fit` method.
    cv : int, cross-validation generator or an iterable, default=5
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
        - None, to use the default 5-fold cross-validation,
        - int, to specify the number of folds in a (Stratified)KFold,
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.
    scoring : str, callable or None, default=None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.
    random_state : int or None, default=None
        Controls the random seed given at the beginning of the algorithm.
    n_jobs : int or None, default=None
        The number of jobs to run in parallel. None means 1 unless in a
        joblib.parallel_backend context. -1 means using all processors.
    importance_getter : str or callable, default='auto'
        If 'auto', uses the feature importance either through a `coef_`
        or `feature_importances_` attributes of estimator.

        Also accepts a string that specifies an attribute name/path
        for extracting feature importance.
        For example, give `regressor_.coef_` in case of
        `~sklearn.compose.TransformedTargetRegressor`  or
        `named_steps.clf.feature_importances_` in case of
        `~sklearn.pipeline.Pipeline` with its last step named `clf`.

        If `callable`, overrides the default feature importance getter.
        The callable is passed with the fitted estimator and the validation set
        (X_val, y_val, estimator) and it should return importance for each feature.
    min_features_to_select : int or float, default=1
        The minimum number of features to select. If float, it represents the
        fraction of features to select.
    init_avg_features_num : float, default=15
        The average number of features to select in the initial pool of solutions.
    init_std_features_num : float, default=5
        The standard deviation of the number of features to select in the initial pool of solutions.
    pool_size : int, default=20
        The number of solutions in the pool.
    n_children_cross_over : int, default=5
        The number of children to create by cross-over.
    is_parent_selection_chance_proportional_to_fitness : bool, default=True
        If True, the probability of selecting a parent is proportional to its fitness. This means
        that the fittest parents are more likely to be selected during crossover.
    n_parents_cross_over : int, default=2
        The number of parents to select in each crossover. More than 2 parents can be selected during
        crossover. In that case, the top features of each parent are combined in a round-robin
        fashion to create a children. The number of features of the children is the average of the
        number of features of the parents.
    n_mutations : int, default=5
        The number of mutations to apply to the pool.
    range_change_n_features_mutation : tuple, default=(-2, 3)
        The range of the number of features to change during mutation. The first element is the
        minimum number of features to change and the second element is the maximum number of features
        to change. The right limit is exclusive.
    range_randomly_swapped_features_mutation : tuple, default=(1, 4)
        The range of the number of features to replace during mutation. The first element is the
        minimum number of features to replace and the second element is the maximum number of
        features to replace. The right limit is exclusive.
    max_generations : int, default=100
        The maximum number of generations to run the genetic algorithm.
    patience : int, default=5
        The number of generations without improvement to wait before stopping the algorithm.
    callbacks : list of callable, default=None
        A list of callables that are called after each generation. Each callable should accept
        the selector and the pool as arguments.
    fitness_function : str or callable, default=rank_mean_test_score_overfit_fitness
        The fitness function to use. Possible string values are: `'mean_test_score'`,
        `'mean_train_score'`, If a callable is passed, it should accept a list of dictionaries where
        each dictionary has the following keys 'features', 'mean_test_score', 'mean_train_score' and
        return a list of floats with the fitness of each element in the pool. Defaults to
        [rank_mean_test_score_overfit_fitness](./#felimination.ga.rank_mean_test_score_fitness)

    Attributes
    ----------

    estimator_ : object
        The fitted estimator.

    support_ : array of shape (n_features,)
        The mask of selected features.

    best_solution_ : dict
        The best solution found by the genetic algorithm. It is a dictionary with the following keys
        - features: list of int
            The features selected for this element.
        - mean_test_score: float
            The mean test score of the element.
        - mean_train_score: float
            The mean train score of the element.
        - train_scores_per_fold: list of float
            The train score of each fold.
        - test_scores_per_fold: list of float
            The test score of each fold.
        - cv_importances: list of array
            The importances of each fold.
        - mean_cv_importances: array
            The mean importances of each fold.

    best_solutions_ : list of dict
        The best solutions found by the genetic algorithm at each generation. Each element is
        defined as in `best_solution_`.

    Examples
    --------

    >>> from felimination.ga import HybridImportanceGACVFeatureSelector
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.linear_model import LogisticRegression
    >>> X, y = make_classification(
        n_samples=sample_size,
        n_features=2,
        n_classes=2,
        n_redundant=0,
        n_clusters_per_class=1,
        random_state=random_state,
    )
    >>> estimator = LogisticRegression(random_state=42)
    >>> selector = selector = HybridImportanceGACVFeatureSelector(
        random_state=random_state,
        init_avg_features_num=2,
        init_std_features_num=1,
    )
    >>> selector = selector.fit(X, y)
    >>> selector.support_
    array([ True,  True,  True,  True,  True, False, False, False, False,
           False])
    """

    _parameter_constraints: dict = {
        "estimator": [HasMethods(["fit"])],
        "min_features_to_select": [
            None,
            Interval(RealNotInt, 0, 1, closed="right"),
            Interval(Integral, 0, None, closed="neither"),
        ],
        "init_avg_features_num": [
            Interval(Real, 0, None, closed="neither"),
        ],
        "init_std_features_num": [
            Interval(Real, 0, None, closed="neither"),
        ],
        "pool_size": [Interval(Integral, 1, None, closed="neither")],
        "n_children_cross_over": [Interval(Integral, 0, None, closed="neither")],
        "is_parent_selection_chance_proportional_to_fitness": [bool],
        "n_parents_cross_over": [Interval(Integral, 2, None, closed="left")],
        "n_mutations": [Interval(Integral, 0, None, closed="neither")],
        "max_generations": [Interval(Integral, 1, None, closed="neither")],
        "patience": [Interval(Integral, 1, None, closed="neither")],
        "importance_getter": [str, callable],
    }

    def __init__(
        self,
        estimator: BaseEstimator | LogisticRegression,
        *,
        cv=5,
        scoring=None,
        random_state=None,
        n_jobs=None,
        importance_getter="auto",
        min_features_to_select=1,
        init_avg_features_num=15,
        init_std_features_num=5,
        pool_size=20,
        is_parent_selection_chance_proportional_to_fitness=True,
        n_children_cross_over=5,
        n_parents_cross_over=2,
        n_mutations=5,
        range_change_n_features_mutation=(-2, 3),
        range_randomly_swapped_features_mutation=(1, 4),
        max_generations=100,
        patience=5,
        callbacks=None,
        fitness_function=rank_mean_test_score_overfit_fitness,
    ) -> None:
        self.estimator = estimator
        self.cv = cv
        self.scoring = scoring
        self.random_state = random_state
        self.n_jobs = n_jobs
        self.importance_getter = importance_getter
        self.min_features_to_select = min_features_to_select
        self.init_avg_features_num = init_avg_features_num
        self.init_std_features_num = init_std_features_num
        self.pool_size = pool_size
        self.n_children_cross_over = n_children_cross_over
        self.is_parent_selection_chance_proportional_to_fitness = (
            is_parent_selection_chance_proportional_to_fitness
        )
        self.n_parents_cross_over = n_parents_cross_over
        self.n_mutations = n_mutations
        self.range_change_n_features_mutation = range_change_n_features_mutation
        self.range_randomly_swapped_features_mutation = (
            range_randomly_swapped_features_mutation
        )
        self.max_generations = max_generations
        self.patience = patience
        self.callbacks = callbacks
        self.fitness_function = fitness_function

    @property
    def best_solution_(self):
        check_is_fitted(self)
        return self.best_solutions_[-1]

    def _evaluate_calculate_importances(
        self, pool, X, y, groups, cv, scorer, **fit_params
    ):
        if effective_n_jobs(self.n_jobs) == 1:
            parallel, func = list, _train_score_get_importance
        else:
            parallel = Parallel(n_jobs=self.n_jobs)
            func = delayed(_train_score_get_importance)

        # Train model, score it and get importances
        # This array has a shape of (n_elements * n_splits, 3)
        # where at each row we have the train score, test score and importances
        # for a given element and split. each group of n_splits rows corresponds
        # to an element in the pool. So for example if we have 2 elements and 3 splits the
        # array will look like:
        # [
        #   [train_score_el1_split1, test_score_el1_split1, importances_el1_split1],
        #   [train_score_el1_split2, test_score_el1_split2, importances_el1_split2],
        #   [train_score_el1_split3, test_score_el1_split3, importances_el1_split3],
        #   [train_score_el2_split1, test_score_el2_split1, importances_el2_split1],
        #   [train_score_el2_split2, test_score_el2_split2, importances_el2_split2],
        #   [train_score_el2_split3, test_score_el2_split3, importances_el2_split3],
        # ]
        scores_importances_1d_array = parallel(
            func(
                self.estimator,
                _select_X_with_features(X, element["features"]),
                y,
                train,
                test,
                scorer,
                self.importance_getter,
                **fit_params,
            )
            for element in pool
            for train, test in cv.split(X, y, groups)
        )
        n_splits = len(list(cv.split(X, y, groups)))
        for i, element in enumerate(pool):
            scores_importances = scores_importances_1d_array[
                i * n_splits : (i + 1) * n_splits
            ]
            train_scores_per_fold = [
                score_importance[0] for score_importance in scores_importances
            ]
            test_scores_per_fold = [
                score_importance[1] for score_importance in scores_importances
            ]
            cv_importances = [
                score_importance[2] for score_importance in scores_importances
            ]
            element["train_scores_per_fold"] = train_scores_per_fold
            element["test_scores_per_fold"] = test_scores_per_fold
            element["cv_importances"] = cv_importances
            element["mean_train_score"] = np.mean(train_scores_per_fold)
            element["mean_test_score"] = np.mean(test_scores_per_fold)
            element["mean_cv_importances"] = np.mean(np.vstack(cv_importances), axis=0)
        return pool

    def _calculate_fitness(self, pool):
        if isinstance(self.fitness_function, str):
            fitness = [el[self.fitness_function] for el in pool]
        else:
            fitness = self.fitness_function(pool)

        return fitness

    def _combine_parents(self, parents):

        sorted_features = [
            np.array(parent["features"])[
                np.argsort(-np.array(parent["mean_cv_importances"]))
            ]
            for parent in parents
        ]
        combined_features = _deduplicate(_roundrobin(*sorted_features))
        n_features = int(
            round(np.mean([len(parent["features"]) for parent in parents]))
        )

        return {
            "features": combined_features[:n_features],
        }

    def _cross_over(self, pool):
        fitness = self._calculate_fitness(pool)
        children = []
        if self.is_parent_selection_chance_proportional_to_fitness:
            parent_selection_proba = fitness / np.sum(fitness)
        else:
            parent_selection_proba = None
        for _ in range(self.n_children_cross_over):
            parents = np.random.choice(
                pool,
                size=self.n_parents_cross_over,
                replace=False,
                p=parent_selection_proba,
            )
            children.append(self._combine_parents(parents))

        return children

    def _mutate(self, pool, all_features):
        mutated_pool = []
        for _ in range(self.n_mutations):
            element = np.random.choice(pool)

            # Randomly increase or decrease the number of features
            number_of_features = max(
                len(element["features"])
                + np.random.randint(*self.range_change_n_features_mutation),
                self.min_features_to_select,
            )

            # Replace the least important features with random features
            n_of_features_to_replace = np.random.randint(
                *self.range_randomly_swapped_features_mutation
            )

            shuffled_all_features = [
                feat for feat in all_features if feat not in element["features"]
            ]
            np.random.shuffle(shuffled_all_features)
            shuffled_all_features = list(shuffled_all_features)

            sorted_features = list(
                np.array(element["features"])[
                    np.argsort(-np.array(element["mean_cv_importances"]))
                ]
            )

            mutated_features = (
                sorted_features[:-n_of_features_to_replace] + shuffled_all_features
            )[:number_of_features]

            mutated_pool.append(
                {
                    "features": mutated_features,
                }
            )
        return mutated_pool

    def _get_support_mask(self):
        check_is_fitted(self)
        return self.support_

    def fit(self, X, y, groups=None, **fit_params):
        """Fit the selector and then the underlying estimator on the selected features.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples.
        y : array-like of shape (n_samples,)
            The target values.
        **fit_params : dict
            Additional parameters passed to the `fit` method of the underlying
            estimator.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        self._validate_params()
        validate_data(
            self,
            X,
            y,
            accept_sparse="csc",
            ensure_min_features=2,
            ensure_all_finite=not get_tags(self.estimator).input_tags.allow_nan,
            multi_output=True,
            dtype=None,
        )

        # Initialization
        cv = check_cv(self.cv, y, classifier=is_classifier(self.estimator))
        scorer = check_scoring(self.estimator, scoring=self.scoring)
        n_features = X.shape[1]
        if self.min_features_to_select is None:
            min_features_to_select = n_features // 2
        elif isinstance(self.min_features_to_select, Integral):  # int
            min_features_to_select = self.min_features_to_select
        else:  # float
            min_features_to_select = int(n_features * self.min_features_to_select)

        if isinstance(X, pd.DataFrame):
            all_features = X.columns.to_list()
        else:
            all_features = list(range(n_features))

        np.random.seed(self.random_state)

        # Create the initial pool of solutions
        pool = [
            {
                "features": list(
                    np.random.choice(
                        all_features,
                        min(
                            max(
                                int(
                                    np.random.normal(
                                        self.init_avg_features_num,
                                        self.init_std_features_num,
                                    )
                                ),
                                min_features_to_select,
                            ),
                            n_features,
                        ),
                        replace=False,
                    )
                ),
            }
            for _ in range(self.pool_size)
        ]

        # Evaluate the initial pool of solutions
        pool = self._evaluate_calculate_importances(
            pool, X, y, groups, cv, scorer, **fit_params
        )
        self.best_solutions_ = []
        for _ in range(1, self.max_generations):
            children = self._cross_over(pool)
            children = self._evaluate_calculate_importances(
                children, X, y, groups, cv, scorer, **fit_params
            )
            pool.extend(children)
            mutations = self._mutate(pool, all_features)
            mutations = self._evaluate_calculate_importances(
                mutations, X, y, groups, cv, scorer, **fit_params
            )
            pool.extend(mutations)
            pool_sorted = [
                element
                for _, element in sorted(
                    zip(self._calculate_fitness(pool), pool),
                    reverse=True,
                    key=itemgetter(0),
                )
            ]
            pool = pool_sorted[: self.pool_size]
            self.best_solutions_.append(pool[0])

            if self.callbacks:
                for callback in self.callbacks:
                    callback(self, pool)

            if len(self.best_solutions_) > self.patience:
                if all(
                    [
                        self.best_solutions_[-1]["features"] == solution["features"]
                        for solution in self.best_solutions_[-self.patience :]
                    ]
                ):
                    break

        self.estimator_ = clone(self.estimator)
        X_remaining_features = _select_X_with_features(
            X, self.best_solution_["features"]
        )
        self.estimator_.fit(X_remaining_features, y, **fit_params)
        self.support_ = np.array(
            [
                True if feature in self.best_solution_["features"] else False
                for feature in all_features
            ]
        )

        return self

    def plot(self, **kwargs):
        """Plot the mean test score and mean train score of the best solution at each generation.

        Parameters
        ----------
        **kwargs : dict
            Additional parameters passed to seaborn.lineplot. For a list
            of possible options, please visit
            [seaborn.lineplot](https://seaborn.pydata.org/generated/seaborn.lineplot.html)  # noqa

        Returns
        -------
        matplotlib.axes.Axes
            The axis where the plot has been plotted.
        """
        data_points_to_plot_long_form = []
        for generation, best_solution in enumerate(self.best_solutions_, start=1):
            for set, scores in zip(
                ["validation", "train"],
                [
                    best_solution["test_scores_per_fold"],
                    best_solution["train_scores_per_fold"],
                ],
            ):
                for score in scores:
                    data_points_to_plot_long_form.append(
                        {"generation": generation, "score": score, "set": set}
                    )
        df_plot = pd.DataFrame(data_points_to_plot_long_form)
        lineplot_kwargs = dict(
            x="generation",
            y="score",
            hue="set",
            markers=True,
            style="set",
            hue_order=["validation", "train"],
            style_order=["validation", "train"],
            seed=self.random_state,
        )
        lineplot_kwargs.update(**kwargs)
        return sns.lineplot(data=df_plot, **lineplot_kwargs)

    @available_if(_estimator_has("predict"))
    def predict(self, X):
        """Reduce X to the selected features and predict using the estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_features]
            The input samples.

        Returns
        -------
        y : array of shape [n_samples]
            The predicted target values.
        """
        check_is_fitted(self)
        return self.estimator_.predict(self.transform(X))

    @available_if(_estimator_has("score"))
    def score(self, X, y, **fit_params):
        """Reduce X to the selected features and return the score of the estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_features]
            The input samples.

        y : array of shape [n_samples]
            The target values.

        **fit_params : dict
            Parameters to pass to the `score` method of the underlying
            estimator.

            .. versionadded:: 1.0

        Returns
        -------
        score : float
            Score of the underlying base estimator computed with the selected
            features returned by `rfe.transform(X)` and `y`.
        """
        check_is_fitted(self)
        return self.estimator_.score(self.transform(X), y, **fit_params)

    @available_if(_estimator_has("decision_function"))
    def decision_function(self, X):
        """Compute the decision function of ``X``.

        Parameters
        ----------
        X : {array-like or sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.

        Returns
        -------
        score : array, shape = [n_samples, n_classes] or [n_samples]
            The decision function of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
            Regression and binary classification produce an array of shape
            [n_samples].
        """
        check_is_fitted(self)
        return self.estimator_.decision_function(self.transform(X))

    @available_if(_estimator_has("predict_proba"))
    def predict_proba(self, X):
        """Predict class probabilities for X.

        Parameters
        ----------
        X : {array-like or sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.

        Returns
        -------
        p : array of shape (n_samples, n_classes)
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
        """
        check_is_fitted(self)
        return self.estimator_.predict_proba(self.transform(X))

    @available_if(_estimator_has("predict_log_proba"))
    def predict_log_proba(self, X):
        """Predict class log-probabilities for X.

        Parameters
        ----------
        X : array of shape [n_samples, n_features]
            The input samples.

        Returns
        -------
        p : array of shape (n_samples, n_classes)
            The class log-probabilities of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
        """
        check_is_fitted(self)
        return self.estimator_.predict_log_proba(self.transform(X))

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        sub_estimator_tags = get_tags(self.estimator)
        tags.target_tags.required = False
        tags.input_tags.allow_nan = sub_estimator_tags.input_tags.allow_nan
        return tags



"""Module with tools to perform feature selection.

This module contains the following classes:

- `FeliminationRFECV`: base class for feature selection.
- `PermutationImportanceRFECV`: recursive feature elimination with
    cross-validation based on permutation importance.
"""

from collections import defaultdict
from numbers import Integral

import numpy as np
import pandas as pd
import seaborn as sns
from joblib import Parallel, delayed, effective_n_jobs
from sklearn.base import BaseEstimator, clone, is_classifier
from sklearn.feature_selection import RFECV
from sklearn.linear_model._logistic import LogisticRegression
from sklearn.metrics import check_scoring
from sklearn.model_selection import check_cv, cross_validate
from sklearn.utils import Bunch
from sklearn.utils._metadata_requests import _routing_enabled, process_routing
from sklearn.utils._param_validation import HasMethods, Interval, RealNotInt
from sklearn.utils._tags import get_tags
from sklearn.utils.validation import check_is_fitted, validate_data

from felimination.importance import PermutationImportance, _train_score_get_importance


class FeliminationRFECV(RFECV):
    """Perform recursive feature elimination with cross-validation
    following scikit-learn standards.

    It has the following differences with RFECV from scikit-learn:

    - It supports an `importance_getter` function that also uses a validation
    set to compute the feature importances. This allows to use importance measures
    like permutation importance or shap.
    - Instead of using Cross Validation to select the number of features, it
    uses cross validation to get a more accurate estimate of the feature
    importances. This means that the number of features to select has to be
    set during initialization, similarly to RFE.
    - When `step` is a float value it is removes a percentage of the number
    of **remaining** features, not total like in RFE/RFECV. This allows to
    drop big chunks of feature at the beginning of the RFE process and to slow
    down towards the end of the process.
    - Has a plotting function
    - Adds information about the number of features selected at each step in the
    attribute `cv_results_`
    - Allows to change the number of features to be selected after fitting.

    Rater than that, it is a copy-paste of RFE, so credit goes to scikit-learn.

    The algorithm of feature selection goes as follows:
    ```
    while n_features > min_features_to_select:
        - The estimator is trained on the selected features and the score is
          computed using cross validation.
        - feature importance is computed for each validation fold on the validation
          set and then averaged.
        - The least important features are pruned.
        - The pruned features are removed from the dataset.
    ```

    Parameters
    ----------
    estimator : ``Estimator`` instance
        A supervised learning estimator with a ``fit`` method.
    step : int or float, default=1
        If greater than or equal to 1, then ``step`` corresponds to the
        (integer) number of features to remove at each iteration.
        If within (0.0, 1.0), then ``step`` corresponds to the percentage
        (rounded down) of **remaining** features to remove at each iteration.
        Note that the last iteration may remove fewer than ``step`` features in
        order to reach ``min_features_to_select``.
    min_features_to_select : int or float, default=None
        The number of features to select. If `None`, half of the features are
        selected. If integer, the parameter is the absolute number of features
        to select. If float between 0 and 1, it is the fraction of the features to
        select.
    cv : int, cross-validation generator or an iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

            - None, to use the default 5-fold cross-validation,
            - integer, to specify the number of folds.
            - :term:`CV splitter`,
            - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        `~sklearn.model_selection.StratifiedKFold` is used. If the
        estimator is a classifier or if ``y`` is neither binary nor multiclass,
        `~sklearn.model_selection.KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.
    scoring : str, callable or None, default=None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.
    verbose : int, default=0
        Controls verbosity of output.
    n_jobs : int or None, default=None
        Number of cores to run in parallel while fitting across folds.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
    importance_getter : str or callable, default='auto'
        If 'auto', uses the feature importance either through a `coef_`
        or `feature_importances_` attributes of estimator.

        Also accepts a string that specifies an attribute name/path
        for extracting feature importance.
        For example, give `regressor_.coef_` in case of
        `~sklearn.compose.TransformedTargetRegressor`  or
        `named_steps.clf.feature_importances_` in case of
        `~sklearn.pipeline.Pipeline` with its last step named `clf`.

        If `callable`, overrides the default feature importance getter.
        The callable is passed with the fitted estimator and the validation set
        (X_val, y_val, estimator) and it should return importance for each feature.


    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        The classes labels. Only available when `estimator` is a classifier.
    estimator_ : ``Estimator`` instance
        The fitted estimator used to select features.
    cv_results_ : dict of ndarrays
        A dict with keys:
        n_features : ndarray of shape (n_subsets_of_features,)
            The number of features used at that step.
        split(k)_test_score : ndarray of shape (n_subsets_of_features,)
            The cross-validation scores across (k)th fold.
        mean_test_score : ndarray of shape (n_subsets_of_features,)
            Mean of scores over the folds.
        std_test_score : ndarray of shape (n_subsets_of_features,)
            Standard deviation of scores over the folds.
        split(k)_train_score : ndarray of shape (n_subsets_of_features,)
            The cross-validation scores across (k)th fold.
        mean_train_score : ndarray of shape (n_subsets_of_features,)
            Mean of scores over the folds.
        std_train_score : ndarray of shape (n_subsets_of_features,)
            Standard deviation of scores over the folds.
    n_features_ : int
        The number of selected features.
    n_features_in_ : int
        Number of features seen during :term:`fit`. Only defined if the
        underlying estimator exposes such an attribute when fit.
    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.
    ranking_ : ndarray of shape (n_features,)
        The feature ranking, such that ``ranking_[i]`` corresponds to the
        ranking position of the i-th feature. Selected (i.e., estimated
        best) features are assigned rank 1.
    support_ : ndarray of shape (n_features,)
        The mask of selected features.
    callbacks : list of callable, default=None
        List of callables to be called at the end of each step of the feature
        selection. Each callable should accept two parameters: the selector
        and the importances computed at that step.

    Examples
    --------
    The following example shows how to retrieve the 5 most informative
    features in the Friedman #1 dataset.

    >>> from felimination.rfe import FeliminationRFECV
    >>> from felimination.importance import PermutationImportance
    >>> from sklearn.datasets import make_friedman1
    >>> from sklearn.svm import SVR
    >>> X, y = make_friedman1(n_samples=50, n_features=10, random_state=0)
    >>> estimator = SVR(kernel="linear")
    >>> selector = selector = FeliminationRFECV(
        estimator,
        step=1,
        cv=5,
        min_features_to_select=5,
        importance_getter=PermutationImportance()
    )
    >>> selector = selector.fit(X, y)
    >>> selector.support_
    array([ True,  True,  True,  True,  True, False, False, False, False,
           False])
    >>> selector.ranking_
    array([1, 1, 1, 1, 1, 6, 3, 4, 2, 5])
    """

    _parameter_constraints: dict = {
        "estimator": [HasMethods(["fit"])],
        "min_features_to_select": [
            None,
            Interval(RealNotInt, 0, 1, closed="right"),
            Interval(Integral, 0, None, closed="neither"),
        ],
        "step": [
            Interval(Integral, 0, None, closed="neither"),
            Interval(RealNotInt, 0, 1, closed="neither"),
        ],
        "verbose": ["verbose"],
        "importance_getter": [str, callable],
    }

    def __init__(
        self,
        estimator,
        *,
        step=1,
        min_features_to_select=1,
        cv=None,
        scoring=None,
        random_state=None,
        verbose=0,
        n_jobs=None,
        importance_getter="auto",
        callbacks=None,
    ) -> None:
        self.random_state = random_state
        self.callbacks = callbacks
        super().__init__(
            estimator,
            min_features_to_select=min_features_to_select,
            cv=cv,
            scoring=scoring,
            n_jobs=n_jobs,
            step=step,
            verbose=verbose,
            importance_getter=importance_getter,
        )

    @staticmethod
    def _select_X_with_remaining_features(X, support, n_features):
        features = np.arange(n_features)[support]
        if isinstance(X, pd.DataFrame):
            feature_names = X.columns[support]
            X_remaining_features = X[feature_names]
        else:
            X_remaining_features = X[:, features]
        return X_remaining_features, features

    def fit(self, X, y, groups=None, **fit_params):
        """Fit the RFE model and then the underlying estimator on the selected features.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples.
        y : array-like of shape (n_samples,)
            The target values.
        **fit_params : dict
            Additional parameters passed to the `fit` method of the underlying
            estimator.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        if _routing_enabled():
            routed_params = process_routing(self, "fit", **fit_params)
        else:
            routed_params = Bunch(estimator=Bunch(fit=fit_params))

        return self._fit(X, y, groups, **routed_params.estimator.fit)

    def _fit(self, X, y, groups, step_score=None, **fit_params):
        self._validate_params()
        validate_data(
            self,
            X,
            y,
            accept_sparse="csc",
            ensure_min_features=2,
            ensure_all_finite=not get_tags(self.estimator).input_tags.allow_nan,
            multi_output=True,
            dtype=None,
        )

        # Initialization
        cv = check_cv(self.cv, y, classifier=is_classifier(self.estimator))
        scorer = check_scoring(self.estimator, scoring=self.scoring)
        n_features = X.shape[1]

        if self.min_features_to_select is None:
            min_features_to_select = n_features // 2
        elif isinstance(self.min_features_to_select, Integral):  # int
            min_features_to_select = self.min_features_to_select
        else:  # float
            min_features_to_select = int(n_features * self.min_features_to_select)

        support_ = np.ones(n_features, dtype=bool)
        ranking_ = np.ones(n_features, dtype=int)

        current_number_of_features = n_features
        self.cv_results_ = defaultdict(list)

        # Elimination
        while current_number_of_features > min_features_to_select:
            # Select remaining features
            X_remaining_features, features = self._select_X_with_remaining_features(
                X, support=support_, n_features=n_features
            )

            if self.verbose > 0:
                print(
                    "Fitting estimator with %d features." % current_number_of_features
                )

            # Train model, score it and get importances
            if effective_n_jobs(self.n_jobs) == 1:
                parallel, func = list, _train_score_get_importance
            else:
                parallel = Parallel(n_jobs=self.n_jobs)
                func = delayed(_train_score_get_importance)

            scores_importances = parallel(
                func(
                    self.estimator,
                    X_remaining_features,
                    y,
                    train,
                    test,
                    scorer,
                    self.importance_getter,
                )
                for train, test in cv.split(X_remaining_features, y, groups)
            )
            train_scores_per_fold = [
                score_importance[0] for score_importance in scores_importances
            ]
            test_scores_per_fold = [
                score_importance[1] for score_importance in scores_importances
            ]
            cv_importances = [
                score_importance[2] for score_importance in scores_importances
            ]
            mean_importances = np.mean(np.vstack(cv_importances), axis=0)
            ranks = np.argsort(mean_importances)

            # for sparse case ranks is matrix
            ranks = np.ravel(ranks)

            if 0.0 < self.step < 1.0:
                step = int(max(1, self.step * current_number_of_features))
            else:
                step = int(self.step)

            # Eliminate the worst features
            threshold = min(step, current_number_of_features - min_features_to_select)

            support_[features[ranks][:threshold]] = False
            ranking_[np.logical_not(support_)] += 1

            # Update cv scores
            for train_or_test, scores_per_fold in zip(
                ["train", "test"], [train_scores_per_fold, test_scores_per_fold]
            ):
                for i, score in enumerate(scores_per_fold):
                    self.cv_results_[f"split{i}_{train_or_test}_score"].append(score)
                self.cv_results_[f"mean_{train_or_test}_score"].append(
                    np.mean(scores_per_fold)
                )
                self.cv_results_[f"std_{train_or_test}_score"].append(
                    np.std(scores_per_fold)
                )
            self.cv_results_["n_features"].append(current_number_of_features)
            if self.callbacks:
                for callback in self.callbacks:
                    callback(self, cv_importances)

            current_number_of_features = np.sum(support_)
        # Set final attributes

        # Estimate performances of final model
        X_remaining_features, features = self._select_X_with_remaining_features(
            X, support=support_, n_features=n_features
        )

        cv_scores = cross_validate(
            self.estimator,
            X_remaining_features,
            y,
            groups=groups,
            scoring=scorer,
            cv=cv,
            n_jobs=self.n_jobs,
            params=fit_params,
            return_train_score=True,
        )
        self.cv_results_["n_features"].append(current_number_of_features)
        # Update cv scores
        for train_or_test in ["train", "test"]:
            scores_per_fold = cv_scores[f"{train_or_test}_score"]
            for i, score in enumerate(scores_per_fold):
                self.cv_results_[f"split{i}_{train_or_test}_score"].append(score)
            self.cv_results_[f"mean_{train_or_test}_score"].append(
                np.mean(scores_per_fold)
            )
            self.cv_results_[f"std_{train_or_test}_score"].append(
                np.std(scores_per_fold)
            )

        if self.callbacks:
            for callback in self.callbacks:
                callback(self, cv_importances)

        X_remaining_features, features = self._select_X_with_remaining_features(
            X, support=support_, n_features=n_features
        )

        self.estimator_ = clone(self.estimator)
        self.estimator_.fit(X_remaining_features, y, **fit_params)

        self.n_features_ = support_.sum()
        self.support_ = support_
        self.ranking_ = ranking_
        self.cv_results_ = dict(self.cv_results_)
        return self

    def set_min_features_to_select(self, min_features_to_select):
        """Changes the number of features to select after fitting.

        The underlying estimator **will not be retrained**. So this method will not
        alter the behavior of predict/predict_proba but it will change the behavior
        of transform and get_feature_names_out.

        Parameters
        ----------
        min_features_to_select : int
            The number of features to select. Must be a value among
            `cv_results_["n_features"]`

        Returns
        -------
        self : object
            Fitted estimator.

        Raises
        ------
        ValueError
            When the number of features to select has not been tried during the
            feature selection procedure.
        """
        check_is_fitted(self)
        if min_features_to_select not in self.cv_results_["n_features"]:
            raise ValueError(
                f"This selector has not been fitted up with {min_features_to_select}, "
                f"please select a value in {set(self.cv_results_['n_features'])} or "
                "refit the selector changing the step parameter of the min_features_to_select"
            )
        support_ = np.zeros_like(self.support_, dtype=bool)
        support_[np.argsort(self.ranking_)[:min_features_to_select]] = True
        self.support_ = support_
        return self

    def plot(self, **kwargs):
        """Plot a feature selection plot with number of features

        Parameters
        ----------
        **kwargs : dict
            Additional parameters passed to seaborn.lineplot. For a list
            of possible options, please visit
            [seaborn.lineplot](https://seaborn.pydata.org/generated/seaborn.lineplot.html)  # noqa

        Returns
        -------
        matplotlib.axes.Axes
            The axis where the plot has been plotted.
        """
        check_is_fitted(self)
        df = pd.DataFrame(self.cv_results_)
        split_score_cols = [col for col in df if "split" in col]
        df_long_form = df[split_score_cols + ["n_features"]].melt(
            id_vars=["n_features"],
            value_vars=split_score_cols,
            var_name="split",
            value_name="score",
        )
        df_long_form["set"] = np.where(
            df_long_form["split"].str.contains("train"), "train", "validation"
        )
        lineplot_kwargs = dict(
            x="n_features",
            y="score",
            hue="set",
            markers=True,
            style="set",
            hue_order=["validation", "train"],
            style_order=["validation", "train"],
            seed=self.random_state,
        )
        lineplot_kwargs.update(**kwargs)
        ax = sns.lineplot(data=df_long_form, **lineplot_kwargs)
        ax.set_xticks(df.n_features)
        return ax


class PermutationImportanceRFECV(FeliminationRFECV):
    """Preset of FeliminationRFECV using permutation importance as importance getter.

    It has the following differences with RFECV from scikit-learn:

    - It supports an `importance_getter` function that also uses a validation
      set to compute the feature importances. This allows to use importance measures
      like permutation importance or shap.
    - Instead of using Cross Validation to select the number of features, it
      uses cross validation to get a more accurate estimate of the feature
      importances. This means that the number of features to select has to be
      set during initialization, similarly to RFE.
    - When `step` is a float value it is removes a percentage of the number
      of **remaining** features, not total like in RFE/RFECV. This allows to
      drop big chunks of feature at the beginning of the RFE process and to slow
      down towards the end of the process.
    - Has a plotting function
    - Adds information about the number of features selected at each step in the
      attribute `cv_results_`
    - Allows to change the number of features to be selected after fitting.

    Rater than that, it is a copy-paste of RFE, so credit goes to scikit-learn.

    The algorithm of feature selection goes as follows:
    ```
    while n_features > min_features_to_select:
        - The estimator is trained on the selected features and the score is
          computed using cross validation.
        - feature importance is computed for each validation fold on the validation
          set and then averaged.
        - The least important features are pruned.
        - The pruned features are removed from the dataset.
    ```

    Parameters
    ----------
    estimator : ``Estimator`` instance
        A supervised learning estimator with a ``fit`` method.
    step : int or float, default=1
        If greater than or equal to 1, then ``step`` corresponds to the
        (integer) number of features to remove at each iteration.
        If within (0.0, 1.0), then ``step`` corresponds to the percentage
        (rounded down) of **remaining** features to remove at each iteration.
        Note that the last iteration may remove fewer than ``step`` features in
        order to reach ``min_features_to_select``.
    min_features_to_select : int or float, default=None
        The number of features to select. If `None`, half of the features are
        selected. If integer, the parameter is the absolute number of features
        to select. If float between 0 and 1, it is the fraction of the features to
        select.
    cv : int, cross-validation generator or an iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 5-fold cross-validation,
        - integer, to specify the number of folds.
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        `~sklearn.model_selection.StratifiedKFold` is used. If the
        estimator is a classifier or if ``y`` is neither binary nor multiclass,
        `~sklearn.model_selection.KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.
    scoring : str, callable or None, default=None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.
    verbose : int, default=0
        Controls verbosity of output.
    n_jobs : int or None, default=None
        Number of cores to run in parallel while fitting across folds.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
    n_repeats : int, default=5
        Number of times to permute a feature.
    random_state : int, RandomState instance, default=None
        Pseudo-random number generator to control the permutations of each
        feature.
        Pass an int to get reproducible results across function calls.
    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights used in scoring.
    max_samples : int or float, default=1.0
        The number of samples to draw from X to compute feature importance
        in each repeat (without replacement).
        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples.
        - If `max_samples` is equal to `1.0` or `X.shape[0]`, all samples
        will be used.
        While using this option may provide less accurate importance estimates,
        it keeps the method tractable when evaluating feature importance on
        large datasets. In combination with `n_repeats`, this allows to control
        the computational speed vs statistical accuracy trade-off of this method.
    callbacks : list of callable, default=None
        List of callables to be called at the end of each step of the feature
        selection. Each callable should accept two parameters: the selector
        and the importances computed at that step.


    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        The classes labels. Only available when `estimator` is a classifier.
    estimator_ : ``Estimator`` instance
        The fitted estimator used to select features.
    cv_results_ : dict of ndarrays
        A dict with keys:
        n_features : ndarray of shape (n_subsets_of_features,)
            The number of features used at that step.
        split(k)_test_score : ndarray of shape (n_subsets_of_features,)
            The cross-validation scores across (k)th fold.
        mean_test_score : ndarray of shape (n_subsets_of_features,)
            Mean of scores over the folds.
        std_test_score : ndarray of shape (n_subsets_of_features,)
            Standard deviation of scores over the folds.
        split(k)_train_score : ndarray of shape (n_subsets_of_features,)
            The cross-validation scores across (k)th fold.
        mean_train_score : ndarray of shape (n_subsets_of_features,)
            Mean of scores over the folds.
        std_train_score : ndarray of shape (n_subsets_of_features,)
            Standard deviation of scores over the folds.
    n_features_ : int
        The number of selected features.
    n_features_in_ : int
        Number of features seen during :term:`fit`. Only defined if the
        underlying estimator exposes such an attribute when fit.
    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.
    ranking_ : ndarray of shape (n_features,)
        The feature ranking, such that ``ranking_[i]`` corresponds to the
        ranking position of the i-th feature. Selected (i.e., estimated
        best) features are assigned rank 1.
    support_ : ndarray of shape (n_features,)
        The mask of selected features.

    Examples
    --------
    The following example shows how to retrieve the 5 most informative
    features in the Friedman #1 dataset.

    >>> from felimination.rfe import PermutationImportanceRFECV
    >>> from sklearn.datasets import make_friedman1
    >>> from sklearn.svm import SVR
    >>> X, y = make_friedman1(n_samples=50, n_features=10, random_state=0)
    >>> estimator = SVR(kernel="linear")
    >>> selector = selector = PermutationImportanceRFECV(
            estimator,
            step=1,
            cv=5,
            min_features_to_select=5,
        )
    >>> selector = selector.fit(X, y)
    >>> selector.support_
    array([ True,  True,  True,  True,  True, False, False, False, False,
           False])
    >>> selector.ranking_
    array([1, 1, 1, 1, 1, 6, 3, 4, 2, 5])
    """

    def __init__(
        self,
        estimator: BaseEstimator | LogisticRegression,
        *,
        step=1,
        min_features_to_select=1,
        cv=None,
        scoring=None,
        verbose=0,
        n_jobs=None,
        n_repeats=5,
        random_state=None,
        sample_weight=None,
        max_samples=1.0,
        callbacks=None,
    ) -> None:
        self.n_repeats = n_repeats
        self.sample_weight = sample_weight
        self.max_samples = max_samples
        super().__init__(
            estimator,
            step=step,
            min_features_to_select=min_features_to_select,
            cv=cv,
            random_state=random_state,
            scoring=scoring,
            verbose=verbose,
            n_jobs=n_jobs,
            callbacks=callbacks,
            importance_getter=PermutationImportance(
                scoring=scoring,
                n_repeats=n_repeats,
                # Better not to do double parallelization
                n_jobs=1,
                random_state=random_state,
                sample_weight=sample_weight,
                max_samples=max_samples,
            ),
        )
